#define HOSTNAME "ESP8266-GarageDoor"

#define TELNET_PORT 23

#define WIFI_SSID ""
#define WIFI_PSK ""
#define WIFI_RETRY_INTERVAL 100 // ms

#define MQTT_SERVER ""
#define MQTT_PORT 1883
#define MQTT_TOPIC "hass/garagedoor"
#define MQTT_USERNAME "homeassistant"
#define MQTT_PASSWORD ""
#define MQTT_RETRY_INTERVAL 500 // ms
#define MQTT_PUBLISH_INTERVAL 1000 // ms

#define ULTRASONIC_DIST_OPEN 70 // cm
#define ULTRASONIC_TRIG_PIN 5 // D1
#define ULTRASONIC_ECHO_PIN 4 // D2
#define ULTRASONIC_TIMEOUT 40000 // µs

#define VERIFICATION_INTERVAL 3000 // ms that the state must stay the same before it is reported

#define LED_PIN 16 // D0

#include <ArduinoOTA.h>
#include <PubSubClient.h>
#include <Ultrasonic.h>

#if defined(ESP8266)
	#include <ESP8266WiFi.h>
#else
	#include <WiFi.h>
#endif

Ultrasonic ultrasonic(ULTRASONIC_TRIG_PIN, ULTRASONIC_ECHO_PIN, ULTRASONIC_TIMEOUT);
WiFiClient wifi, telnet;
WiFiServer telnetServer(TELNET_PORT);
PubSubClient mqtt(wifi);
int distance;

void setupWifi() {
	if (WiFi.status() == WL_CONNECTED) {
		return;
	}
	
	Serial.print("[WiFi] Connecting to ");
	Serial.print(WIFI_SSID);

	WiFi.hostname(HOSTNAME);
	WiFi.mode(WIFI_STA);
	WiFi.begin(WIFI_SSID, WIFI_PSK);

	while (WiFi.status() != WL_CONNECTED) {
		delay(WIFI_RETRY_INTERVAL);
		Serial.print(".");
	}

	Serial.println();
	Serial.print("[WiFi] Connected as ");
	Serial.println(WiFi.localIP());
}

void setupTelnet() {
	if (!(telnet = telnetServer.available())) {
		return;
	}

	telnet.print("mqtt_connected = ");
	telnet.println(mqtt.connected());
	
	telnet.print("distance_cm = ");
	telnet.println(distance);
	
	telnet.print("door_open = ");
	telnet.println(isDoorOpen());
	
	telnet.flush();
	telnet.stop();
}

void setupMqtt() {
	if (mqtt.connected()) {
		return;
	}
	
	mqtt.setServer(MQTT_SERVER, MQTT_PORT);
	
	Serial.print("[MQTT] Connecting to ");
    Serial.print(MQTT_SERVER);
    Serial.print(":");
    Serial.print(MQTT_PORT);
    Serial.print(" as ");
    Serial.print(MQTT_USERNAME);

    mqtt.connect(HOSTNAME, MQTT_USERNAME, MQTT_PASSWORD);

    if (mqtt.connected()) {
        Serial.println();
        Serial.println("[MQTT] Connected");
    } else {
        Serial.println();
        Serial.print("[MQTT] Error connecting: ");
        Serial.println(mqtt.state());   
    }
}

void setup() {
	Serial.begin(9600);
	telnetServer.begin();
	delay(10);

	ArduinoOTA.begin();
	ArduinoOTA.setHostname(HOSTNAME);

	pinMode(LED_PIN, OUTPUT);
	digitalWrite(LED_PIN, LOW);

	setupWifi();
	setupMqtt();

	// Turn on the LED for 1 second when setup is complete
	digitalWrite(LED_PIN, HIGH);
	delay(1000);
	digitalWrite(LED_PIN, LOW);
}

bool isDoorOpen() {
    distance = ultrasonic.distanceRead();
	return distance <= ULTRASONIC_DIST_OPEN;
}

void loop() {
	setupWifi();
	setupTelnet();
	setupMqtt();
	ArduinoOTA.handle();
	
	bool state = isDoorOpen();
	delay(VERIFICATION_INTERVAL);
	bool stateVerify = isDoorOpen();
	
	if (state == stateVerify) {
		digitalWrite(LED_PIN, !state);
	
        Serial.print("[loop] Distance = ");
        Serial.print(distance);
        Serial.println(" cm");
		Serial.print("[loop] Door is ");
		Serial.println(state ? "Open" : "Closed");
		
		mqtt.publish(MQTT_TOPIC, state ? "1" : "0");
		delay(MQTT_PUBLISH_INTERVAL);
	}
}

